import functools
from datetime import datetime

_last_result_time = None
_last_result_value = None
def ttl_cache(ttl=datetime.timedelta(hours=1)):
    def wrap(func):
        cache = {}
        @functools.wraps(func)
        def wrapped(*args, **kw):
            now = datetime.datetime.now()
            # see lru_cache for fancier alternatives
            key = tuple(args), frozenset(kw.items())
            if key not in cache or now - cache[key][0] > ttl:
                value = func(*args, **kw)
                cache[key] = (now, value)
            return cache[key][1]
        return wrapped
    return wrap


import types

def staticclass(cls):
    def init(*args, **kwargs):
        raise Exception("Cannot make an instance of a static class.")
    cls.__init__ = init
    for name, value in vars(cls).items():
        if isinstance(value, types.FunctionType):
            setattr(cls, name, staticmethod(value))
    return cls


import traceback
import sys

class TracePrints(object):
    def __init__(self):
        self.stdout = sys.stdout
        self.ignore_print = False
    def write(self, s):
        if self.ignore_print:
            self.ignore_print = False
            return
        # self.stdout.write("Writing %r\n" % s)
        stack = traceback.format_stack()
        last_line = ""
        for line in stack:
            if "traceback.format_stack" in line:
                break
            last_line = line
        self.stdout.write(last_line)
        self.stdout.write(s + "\n---------\n")
        self.ignore_print = True


from datetime import datetime, timezone
from dateutil.parser import parse as parse_time


def convert_from_utc_epoch(time, tz_aware=True):
    if tz_aware:
        return datetime.utcfromtimestamp(time).replace(tzinfo=timezone.utc)
    else:
        return datetime.utcfromtimestamp(time)

def convert_to_utc_datetime(time):
    d = time
    if type(time) == str:
        d = parse_time(time)
    if not d.tzinfo:
        return d.replace(tzinfo=timezone.utc)
    elif d.tzinfo == timezone.utc:
        return d
    else:
        return convert_from_utc_epoch(d.timestamp())

def convert_to_utc_epoch(time):
    return convert_to_utc_datetime(time).timestamp()

def current_utc_datetime():
    return datetime.utcnow().replace(tzinfo=timezone.utc)

def current_utc_epoch():
    return current_utc_datetime().timestamp()



import itertools
from enum import Enum
from get_docker_secret import get_docker_secret


def inherit_attribute(name, f):
    def decorator(cls):
        old_value = getattr(cls, name)
        new_value = f([getattr(base, name) for base in cls.__bases__ if hasattr(base, name)], old_value)
        setattr(cls, name, new_value)
        return cls
    return decorator

def merge_class_val(base_values, my_value = []):
    chain_object = itertools.chain.from_iterable(base_values)
    return list(chain_object) + my_value

def merge_inheritance(name):
    return inherit_attribute(name, merge_class_val)

class CaseInsensitiveEnum(Enum):
    @classmethod
    def _missing_(cls, value):
        if value is None:
            return super()._missing_(value)
        for member in cls:
            if member.name.upper() == value.upper():
                return member
        super()._missing_(value)

IS_PRODUCTION = None
def is_production():
    global IS_PRODUCTION
    if IS_PRODUCTION is None:
        IS_PRODUCTION = get_secret_helper('ENVIRONMENT', 'development') != 'development'
    return IS_PRODUCTION

class MissingSecretException(Exception):
    pass

def get_secret_helper(key, *args, **kwargs):
    s = get_docker_secret(key, default=None)
    if not s:
        if 'default' in kwargs:
            return kwargs['default']
        elif len(args) > 0:
            return args[0]
        elif is_production():
            s = None
        else:
            raise MissingSecretException
    return s

def get_secret_helper_bool(key, default=None, raise_exception=True):
    v = get_secret_helper(key, None)
    if v is None:
        if default is not None or not raise_exception or is_production():
            return default
        else:
            raise MissingSecretException
    else:
        return v.upper() == 'TRUE'


def get_attr_or_value(obj_or_dict, name_key, default=None, throw_error=False):
    try:
        if isinstance(obj_or_dict, dict):
            return obj_or_dict[name_key]
        else:
            return getattr(obj_or_dict, name_key)
    except Exception as e:
        if throw_error:
            raise e
        else:
            return default